#
# Collective Knowledge (platform - GPU)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Detect GPU

def detect(i):
    """
    Input:  {
              (target)               - if specified, use info from 'machine' module

              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)
              (skip_device_init)     - if 'yes', do not initialize device
              (print_device_info)    - if 'yes', print extra device info

              (skip_info_collection) - if 'yes', do not collect info (particularly for remote)

              (skip_print_os_info)   - if 'yes', do not print OS info

              (exchange)             - if 'yes', exchange info with some repo (by default, remote-ck)
              (share)                - the same as 'exchange'
              (exchange_repo)        - which repo to record/update info (remote-ck by default)
              (exchange_subrepo)     - if remote, remote repo UOA
              (exchange_locally)     - if 'yes', exchange locally

              (extra_info)           - extra info about author, etc (see add from CK kernel)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              features = {
                gpu          - GPU features (properties), unified
                gpu_misc     - assorted GPU features (properties), platform dependent
              }
            }

    """

    import os

    o=i.get('out','')

    oo=''
    if o=='con': oo=o

    # Check if target
    if i.get('target','')!='':
       r=ck.access({'action':'init',
                    'module_uoa':cfg['module_deps']['machine'],
                    'input':i})
       if r['return']>0: return r

    device_cfg=i.get('device_cfg',{})

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    sic=i.get('skip_info_collection','')
    sdi=i.get('skip_device_init','')
    pdv=i.get('print_device_info','')

    ex=i.get('exchange','')
    if ex=='': ex=i.get('share','')

    einf=i.get('extra_info','')
    if einf=='': einf={}

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=oo
    if i.get('skip_print_os_info','')=='yes': ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.cpu']
    ii['print_device_info']='no'
    ii['skip_print_os_info']='yes'
    rr=ck.access(ii) # DO NOT USE rr further - will be reused as return !
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tosd.update(device_cfg.get('update_target_os_dict',{}))

    tbits=tosd.get('bits','')

    tdid=rr['device_id']

    # Some params
    ro=tosd.get('redirect_stdout','')
    re=tosd.get('redirect_stderr','')
    remote=tosd.get('remote','')
    remote_ssh=tosd.get('remote_ssh','')
    win=tosd.get('windows_base','')
    mac=tosd.get('macos','')
    unix=win!='yes' and mac!='yes'

    stdirs=tosd.get('dir_sep','')

    dv=''
    if tdid!='': dv=' -s '+tdid

    # Init
    prop={}
    prop_all={}

    target_gpu_name=''
    target_gpu_vendor=''

    # Get info about GPU on Android ######################################################
    if remote=='yes' and remote_ssh!='yes':
       # Get all params
       params={}

       rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
       if rx['return']>0: return rx
       fn=rx['file_name']

       # Get GPU
       x=tosd.get('adb_dumpsys','').replace('$#device#$',dv)
       x=x.replace('$#category#$','SurfaceFlinger')
       x=x.replace('$#redirect_stdout#$', ro)
       x=x.replace('$#output_file#$', fn)

       if o=='con' and pdv=='yes':
          ck.out('')
          ck.out('Receiving all parameters:')
          ck.out('  '+x)

       rx=os.system(x)
       if rx!=0:
          if o=='con':
             ck.out('')
             ck.out('Non-zero return code :'+str(rx)+' - likely failed')
       else:
          # Read and parse file
          rx=ck.load_text_file({'text_file':fn, 'split_to_list':'yes', 'delete_after_read':'yes'})
          if rx['return']>0: return rx
          ll=rx['lst']

          for s in ll:
              s1=s.strip()
              q2=s1.find('GLES: ')
              if q2>=0:
                 x=s1[6:].strip().split(',')

                 if len(x)>0: 
                    target_gpu_vendor=x[0].strip()
                    target_gpu_name+=target_gpu_vendor
                 if len(x)>1: target_gpu_name+=' '+x[1].strip()

                 prop['name']=target_gpu_name
                 prop['vendor']=target_gpu_vendor
                 prop['possibly_related_cpu_name']=''

                 break
    else:
       if win=='yes':
          r=ck.access({'action':'get_from_wmic',
                       'module_uoa':cfg['module_deps']['platform'],
                       'group':'cpu',
                       'remote_shell':tosd.get('remote_shell','').replace('$#device#$',dv),
                       'remote_shell_end':tosd.get('remote_shell_end','')})
          if r['return']>0: return r
          info_cpu=r['dict']

          target_cpu=info_cpu.get('Name','')

          r=ck.access({'action':'get_from_wmic',
                       'module_uoa':cfg['module_deps']['platform'],
                       'cmd':'path Win32_VideoController get Name',
                       'remote_shell':tosd.get('remote_shell','').replace('$#device#$',dv),
                       'remote_shell_end':tosd.get('remote_shell_end','')})
          if r['return']>0: return r
          target_gpu_name=r['value']

          x=target_gpu_name.split(' ')
          if len(x)>0:
             target_gpu_vendor=x[0].strip()

          prop['name']=target_gpu_name
          prop['vendor']=target_gpu_vendor
          prop['possibly_related_cpu_name']=target_cpu

       elif mac=='yes':
          cmd=['system_profiler', 'SPDisplaysDataType']

          if remote_ssh=='yes':
             cmd=tosd['remote_shell']+' system_profiler SPDisplaysDataType '+tosd.get('remote_shell_end','')

          r=ck.access({'action':'run_and_get_stdout',
                       'module_uoa':cfg['module_deps']['os'],
                       'cmd': cmd})
          if r['return']>0: return r

          target_gpu_name=''
          for line in r['stdout'].splitlines():
            if ':' in line:
              left, right = line.split(':', 1)
              left = left.strip().lower()
              if left == 'chipset model':
                target_gpu_name = right.strip()
                break

          x=target_gpu_name.split(' ')
          target_gpu_vendor=''
          if len(x)>0:
             target_gpu_vendor=x[0].strip()

          prop['name']=target_gpu_name
          prop['vendor']=target_gpu_vendor

       else:
          # Get devices via lspci (if supported)
          target_gpu_name=''

          rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
          if rx['return']>0: return rx
          fn=rx['file_name']

          rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
          if rx['return']>0: return rx
          fn1=rx['file_name']

          x='lspci'

          if remote_ssh=='yes':
             x=tosd['remote_shell']+x+tosd.get('remote_shell_end','')

          x+=' '+ro+' '+fn+' '+re+' '+fn1

          if o=='con':
             ck.out('')
             ck.out('Executing: '+x)

          rx=os.system(x)

          if os.path.isfile(fn1):
             os.remove(fn1)

          if rx==0:
             # Read and parse file
             rx=ck.load_text_file({'text_file':fn, 'split_to_list':'yes', 'delete_after_read':'yes'})
             if rx['return']==0: 
                ll=rx['lst']

                for q in ll:
                    x1=q.find('VGA ')
                    if x1>=0:
                       x2=q.find(':', x1+1)
                       if x2>=0:
                          target_gpu_name=q[x2+1:].strip()
                          break

          x=target_gpu_name.split(' ')
          if len(x)>0:
             target_gpu_vendor=x[0].strip()

          prop['name']=target_gpu_name
          prop['vendor']=target_gpu_vendor

    if o=='con' and prop.get('name','')!='':
       ck.out('')
       ck.out('GPU name:   '+prop.get('name',''))
       ck.out('GPU vendor: '+prop.get('vendor',''))

    # Check frequency via script
    if unix=='yes':
       rx=ck.gen_tmp_file({'prefix':'tmp-ck-'})
       if rx['return']>0: return rx
       fn=rx['file_name']

       cmd=tosd.get('script_get_gpu_frequency','')
       if cmd!='':
          cmd+=' '+ro+fn

          # Check path to scripts from env
          path_to_scripts=''

          pi_uoa=os.environ.get('CK_PLATFORM_INIT_UOA','')
          if pi_uoa=='':
             dcfg={}
             ii={'action':'load',
                 'module_uoa':cfg['module_deps']['cfg'],
                 'data_uoa':cfg['lcfg_uoa']}
             r=ck.access(ii)
             if r['return']>0 and r['return']!=16: return r
             if r['return']!=16:
                pi_key=tosx
                if remote=='yes' and tdid!='': pi_key+='-'+tdid

                dcfg=r['dict']
                pi_uoa=dcfg.get('platform_init_uoa',{}).get(pi_key,'')

          if pi_uoa!='' and remote!='yes':
             rx=ck.access({'action':'find',
                           'module_uoa':cfg['module_deps']['platform.init'],
                           'data_uoa':pi_uoa})
             if rx['return']==0:
                path_to_scripts=rx['path']

          if path_to_scripts=='':
             path_to_scripts=tosd.get('path_to_scripts','')

          if path_to_scripts!='': cmd=path_to_scripts+stdirs+cmd

          if remote=='yes':
             # Execute script
             cmd=tosd.get('remote_shell','').replace('$#device#$',dv)+' '+cmd+tosd.get('remote_shell_end','')

          if o=='con':
             ck.out('')
             ck.out('Trying to read GPU frequency:')
             ck.out('  '+cmd)

          rx=os.system(cmd)
          if rx!=0:
             if o=='con':
                ck.out('')
                ck.out('Non-zero return code :'+str(rx)+' - likely failed')
                ck.out('')
          else:
             # Read and parse file
             rx=ck.load_text_file({'text_file':fn, 'split_to_list':'yes', 'delete_after_read':'yes'})
             if rx['return']>0: return rx
             ll=rx['lst']

             cur_freq=''
             freqs=[]

             jl=len(ll)
             for j in range(0,jl):
                 s=ll[j]
                 if s.lower().startswith('*** current gpu frequency:'):
                    if (j+1)<jl:
                       cur_freq=ll[j+1]

                 if s.lower().startswith('*** available gpu frequencies:'):
                    while s!='' and j<jl:
                       j+=1
                       if j<jl:
                          s=ll[j]
                          if s!='':
                             freqs.append(s)
                    break

             prop['current_freq']=cur_freq
             prop['all_freqs']=freqs

             if o=='con' and cur_freq!='':
                ck.out('')
                ck.out('Current GPU frequency:')
                ck.out('  '+str(cur_freq))
                if len(freqs)>0:
                   ck.out('')
                   ck.out('All frequencies:')
                   for q in freqs:
                       ck.out(' '+q)

    fuoa=''
    fuid=''

    # Exchanging info #################################################################
    if ex=='yes':

       er=i.get('exchange_repo','')
       esr=i.get('exchange_subrepo','')
       el=i.get('exchange_locally','')
       if el!='yes' and er=='': 
          er=ck.cfg['default_exchange_repo_uoa']
          esr=ck.cfg['default_exchange_subrepo_uoa']

       if o=='con':
          ck.out('')
          ck.out('Exchanging information with '+er+' repository ...')

       xn=prop.get('name','')
       if xn=='':
          # Check if exists in configuration

          dcfg={}
          ii={'action':'load',
              'module_uoa':cfg['module_deps']['cfg'],
              'data_uoa':cfg['cfg_uoa']}
          r=ck.access(ii)
          if r['return']>0 and r['return']!=16: return r
          if r['return']!=16:
             dcfg=r['dict']

          dx=dcfg.get('platform_gpu_name',{}).get(tos,{})
          x=tdid
          if x=='': x='default'
          xn=dx.get(x,'')

          if (xn=='' and o=='con'):
             r=ck.inp({'text':'Enter your GPU name (for example, Arm Mali-T860) or leave empty if you don\'t know it: '})
             xxn=r['string'].strip()
             if xxn=='': 
                xxn='-'

             if xxn!=xn:
                xn=xxn

                if 'platform_gpu_name' not in dcfg: dcfg['platform_gpu_name']={}
                if tos not in dcfg['platform_gpu_name']: dcfg['platform_gpu_name'][tos]={}
                dcfg['platform_gpu_name'][tos][x]=xn

                ii={'action':'update',
                    'module_uoa':cfg['module_deps']['cfg'],
                    'data_uoa':cfg['cfg_uoa'],
                    'dict':dcfg}
                r=ck.access(ii)
                if r['return']>0: return r

          if xn=='':
             return {'return':1, 'error':'can\'t exchange information where main name is empty'}

          ixn=xn.find(' ')
          if ixn>0: 
             xx=xn[:ixn].strip()
             prop['vendor']=xx
             xn=xn[ixn+1:].strip()

          prop['name']=xn

       ii={'action':'exchange',
           'module_uoa':cfg['module_deps']['platform'],
           'sub_module_uoa':work['self_module_uid'],
           'repo_uoa':er,
           'data_name':prop.get('name',''),
           'extra_info':einf,
           'all':'yes',
           'dict':{'features':prop}} # Later we should add more properties from prop_all,
                                     # but should be careful to remove any user-specific info
       if esr!='': ii['remote_repo_uoa']=esr
       r=ck.access(ii)
       if r['return']>0: return r

       fuoa=r.get('data_uoa','')
       fuid=r.get('data_uid','')

       prop=r['dict'].get('features',{})

       if o=='con' and r.get('found','')=='yes':
          ck.out('  GPU CK entry already exists ('+fuid+') - loading latest meta (features) ...')

    rr={'return':0, 'features':{}}

    rr['features']['gpu']=prop
    rr['features']['gpu_misc']=prop_all

    if fuoa!='' or fuid!='':
       rr['features']['gpu_uoa']=fuoa
       rr['features']['gpu_misc_uid']=fuid

    return rr

##############################################################################
# set frequency

def set_freq(i):
    """
    Input:  {
              (host_os)              - host OS (detect, if omitted)
              (os) or (target_os)    - OS module to check (if omitted, analyze host)

              (device_id)            - device id if remote (such as adb)

              (value) = "max" (default)
                        "min"
                        "ondemand"
                        int value
              (env)                  - current env to specialize freq (CK_GPU*)...
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    v=i.get('value','')
    if v=='': v='max'

    # Various params
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    if tos=='': tos=i.get('os','')
    tdid=i.get('device_id','')

    env=i.get('env',{})

    # Get OS info
    import copy
    ii=copy.deepcopy(i)
    ii['out']=''
    ii['action']='detect'
    ii['module_uoa']=cfg['module_deps']['platform.os']
    ii['skip_info_collection']='yes'
    ii['skip_device_init']='yes'
    rr=ck.access(ii)
    if rr['return']>0: return rr

    hos=rr['host_os_uid']
    hosx=rr['host_os_uoa']
    hosd=rr['host_os_dict']

    tos=rr['os_uid']
    tosx=rr['os_uoa']
    tosd=rr['os_dict']

    tbits=tosd.get('bits','')

    envtsep=tosd.get('env_separator','')
    etset=tosd.get('env_set','')

    tdid=rr['device_id']

    dir_sep=tosd.get('dir_sep','')

    remote=tosd.get('remote','')
    rs=tosd.get('remote_shell','')
    rse=tosd.get('remote_shell_end','')

    # Prepare scripts
    cmd=''
    if v=='min':
       cmd=tosd.get('script_set_min_gpu_freq','')
    elif v=='max':
       cmd=tosd.get('script_set_max_gpu_freq','')
    elif v=='ondemand':
       cmd=tosd.get('script_set_ondemand_gpu_freq','')
    else:
       cmd=tosd.get('script_set_gpu_freq','').replace('$#freq#$',str(v))

    if cmd!='':
       # Check path to scripts from env
       path_to_scripts=''

       pi_uoa=os.environ.get('CK_PLATFORM_INIT_UOA','')
       if pi_uoa=='':
          dcfg={}
          ii={'action':'load',
              'module_uoa':cfg['module_deps']['cfg'],
              'data_uoa':cfg['lcfg_uoa']}
          r=ck.access(ii)
          if r['return']>0 and r['return']!=16: return r
          if r['return']!=16:
             pi_key=tosx
             if remote=='yes' and tdid!='': pi_key+='-'+tdid

             dcfg=r['dict']
             pi_uoa=dcfg.get('platform_init_uoa',{}).get(pi_key,'')

       if pi_uoa!='' and remote!='yes':
          rx=ck.access({'action':'find',
                        'module_uoa':cfg['module_deps']['platform.init'],
                        'data_uoa':pi_uoa})
          if rx['return']==0:
             path_to_scripts=rx['path']

       if path_to_scripts=='':
          path_to_scripts=tosd.get('path_to_scripts','')

       if path_to_scripts!='':
          cmd=path_to_scripts+dir_sep+cmd

       # Add env
       xcmd=''
       for k in sorted(env):
           if k.startswith('CK_GPU'):
              v=str(env[k])

              if xcmd!='': xcmd+=envtsep
              xcmd+=etset+' '+k+'='+v

       if xcmd!='': 
          cmd=xcmd+envtsep+cmd

       if o=='con':
          ck.out('')
          ck.out('CMD to set GPU frequency:')
          ck.out('  '+cmd)

       # Get all params
       if remote=='yes':
          dv=''
          if tdid!='': dv=' -s '+tdid

          x=rs.replace('$#device#$',dv)
          # Support backwards compatibility if double quote not found in rs and rse
          x+=' '+('"' if rs.find('"') == -1 else '')+' '+cmd+' '+rse+('"' if rse.find('"') == -1 else '')

          rx=os.system(x)
          if rx!=0:
             if o=='con':
                ck.out('')
                ck.out('Non-zero return code :'+str(rx)+' - likely failed')

       else:
             rx=os.system(cmd)
             if rx!=0:
                if o=='con':
                   ck.out('')
                   ck.out('  Warning: setting frequency possibly failed - return code '+str(rx))

    return {'return':0}

##############################################################################
# viewing entries as html

def show(i):
    """
    Input:  {
              data_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """


    h='<h2>GPUs participating in crowd-tuning</h2>\n'

    h+='<i>Reuse/extend <a href="https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.gpu">CK JSON meta information</a> of these GPUs using "ck pull repo:ck-crowdtuning-platforms" ...</i><br><br>\n'

    h+='<table class="ck_table" border="0" cellpadding="6" cellspacing="0">\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']

    h+=' <tr style="background-color:#cfcfff;">\n'
    h+='  <td><b>\n'
    h+='   #\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Vendor\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   Name\n'
    h+='  </b></td>\n'
    h+='  <td><b>\n'
    h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':">CK UID</a>\n'
    h+='  </b></td>\n'
    h+=' </tr>\n'

    ruoa=i.get('repo_uoa','')
    muoa=work['self_module_uoa']
    duoa=i.get('data_uoa','')

    r=ck.access({'action':'search',
                 'module_uoa':muoa,
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'add_info':'yes',
                 'add_meta':'yes'})
    if r['return']>0: 
       return {'return':0, 'html':'Error: '+r['error']}

    lst=r['lst']

    num=0
    for q in sorted(lst, key = lambda x: (x.get('meta',{}).get('features',{}).get('vendor','').upper(), \
                                          x.get('meta',{}).get('features',{}).get('name','').upper())):

        num+=1

        duoa=q['data_uoa']
        duid=q['data_uid']

        meta=q['meta']
        ft=meta.get('features',{})
        
        vendor=ft.get('vendor','')
        name=ft.get('name','')

        h+=' <tr>\n'
        h+='  <td valign="top">\n'
        h+='   '+str(num)+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+vendor+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   '+name+'\n'
        h+='  </td>\n'
        h+='  <td valign="top">\n'
        h+='   <a href="'+url0+'wcid='+work['self_module_uoa']+':'+duid+'">'+duid+'</a>\n'
        h+='  </td>\n'
        h+=' </tr>\n'


    h+='</table><br><br>\n'

    return {'return':0, 'html':h}

##############################################################################
# browse platform.gpus participated in experiment crowdsourcing (crowd-benchmarking and crowd-tuning)

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # TBD: should calculate url
    url='https://github.com/ctuning/ck-crowdtuning-platforms/tree/master/platform.gpu'

    import webbrowser
    webbrowser.open(url)

    import time
    time.sleep(3)

    url='http://cknowledge.org/repo/web.php?action=index&module_uoa=wfe&native_action=show&native_module_uoa=platform.gpu'

    import webbrowser
    webbrowser.open(url)

    return {'return':0}



#
# Collective Knowledge (Universal predictive model)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Use model (universal)

def use(i):
    """

    Input:  {
              features                                - Features [f1,f2,...]
              model_module_uoa                        - model module
              model_name                              - model name
              model_file                              - model file

              (model_data_uoa)                        - get model from this data_uoa (container: model_module_uoa)
              (model_repo_uoa)                        - use this repo to record model

            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              prediction                              - [value]
            }

    """

    import os

    o=i.get('out','')
    i['out']=''

    mduoa=i.get('model_data_uoa','')
    mruoa=i.get('model_repo_uoa','')

    if mduoa=='':
       mmuoa=i['model_module_uoa']
       mn=i['model_name']
       mf=i['model_file']
    else:
       r=ck.access({'action':'load',
                    'module_uoa':work['self_module_uid'],
                    'data_uoa':mduoa,
                    'repo_uoa':mruoa})
       if r['return']>0: return r
       p=r['path']
       d=r['dict']

       mmuoa=d['model_module_uoa']
       mn=d['model_name']
       mf=d['model_file']
       
       mf=os.path.join(p,mf)

    # Get table through experiment module for features
    features=i['features']
    ftable=[features]

    # Calling model
    ii={'action':'validate',
        'module_uoa':mmuoa,
        'model_name':mn,
        'model_file':mf,
        'features_table': ftable,
        'model_params':i.get('model_params',{}),
        'out':o
       }
    r=ck.access(ii)
    if r['return']>0: return r

    prediction=r['prediction_table'][0]

    if o=='con':
       ck.out('Features:')
       for ft in range(0, len(features)):
           ck.out('  ft'+str(ft)+' -> '+str(features[ft]))
       ck.out('Prediction:')
       ck.out('  '+str(prediction))

    return {'return':0, 'prediction':prediction}

##############################################################################
# build model (universal)

def build(i):
    """

    Input:  {
              Use data for modeling directly
                (ftable)                                - Feature table
                                                            [ [f1,f2,...], [f1,f2,...], ...]
                (fkeys)                                 - Key names
                                                            [ "f1 name", "f2 name", ...]

                (ctable)                                - Outcome to model
                                                            [ [value], [value], ... ]
                (ckeys)                                 - Key names
                                                            [ "value name" ]
              
              OR select entries:

                (repo_uoa) or (experiment_repo_uoa)     - can be wild cards
                (remote_repo_uoa)                       - if remote access, use this as a remote repo UOA
                (module_uoa) or (experiment_module_uoa) - can be wild cards
                (data_uoa) or (experiment_data_uoa)     - can be wild cards

                (repo_uoa_list)                         - list of repos to search
                (module_uoa_list)                       - list of module to search
                (data_uoa_list)                         - list of data to search

                (search_dict)                           - search dict
                (ignore_case)                           - if 'yes', ignore case when searching

                (features_flat_keys_list)               - list of flat keys to extract from points into table
                                                          (order is important: for example, for plot -> X,Y,Z)
                (features_flat_keys_ext)                - extension of keys with wildcard * to extract multiple keys 
                                                          (for example, ##features#) - useful when many
                (features_flat_keys_desc)               - list of flat key descriptions (not stable!)

                (characteristics_flat_keys_list)        - list of flat keys to extract from points into table
                                                          (order is important: for example, for plot -> X,Y,Z)
                (characteristics_flat_keys_index)       - add all flat keys starting from this index 
                                                          (for example, ##features#)

              Model:
                model_module_uoa                        - model module
                model_name                              - model name
                (model_params)                          - dict with model params
                (model_file)                            - model output file, otherwise generated as tmp file
                (keep_temp_files)                       - if 'yes', keep temp files 
                (remove_points_with_none)               - if 'yes', remote points with None
                (caption)                               - add caption to graphs, if needed

                (model_data_uoa)                       - record to this data_uoa (container: model_module_uoa)
                (model_repo_uoa)                       - use this repo to record model

              (csv_file)                                - if !='', only record prepared table to CSV ...
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy
    import os

    o=i.get('out','')
    i['out']=''

    ktf=i.get('keep_temp_files','')

    cap=i.get('caption','')

    rpwn=i.get('remove_points_with_none','')

    cf=i.get('csv_file','')

    # Get table through experiment module for features
    ffkl=i.get('features_flat_keys_list',[])
    fdesc=i.get('features_flat_keys_desc',{})

    ffke=i.get('features_flat_keys_ext','')
    if ffke!='':
       ffkl1=[]
       for q in ffkl:
           q+=ffke
           ffkl1.append(q)
       ffkl=ffkl1

       fdesc1={}
       for q in fdesc:
           fdesc1[q+ffke]=fdesc[q]
       fdesc=fdesc1

    ftable=i.get('ftable',[])
    fkeys=i.get('fkeys',[])
    if len(ftable)==0:
       iif=copy.deepcopy(i)
       iif['action']='get'
       iif['module_uoa']=cfg['module_deps']['experiment']
       iif['flat_keys_list']=ffkl
       iif['flat_keys_index']=i.get('features_flat_keys_index','')
       r=ck.access(iif)
       if r['return']>0: return r
       ftable=r['table'].get('0',[])
       fkeys=r['real_keys']

       if len(ftable)==0:
          return {'return':1, 'error':'no points found'}

    # Get table through experiment module for characteristics
    ctable=i.get('ctable',[])
    ckeys=i.get('ckeys',[])
    if len(ctable)==0:
       iic=copy.deepcopy(i)
       iic['action']='get'
       iic['module_uoa']=cfg['module_deps']['experiment']
       iic['flat_keys_list']=i.get('characteristics_flat_keys_list',[])
       iic['flat_keys_index']=i.get('characteristics_flat_keys_index','')
       r=ck.access(iic)
       if r['return']>0: return r
       ctable=r['table'].get('0',[])
       ckeys=r['real_keys']

       if len(ctable)==0:
          return {'return':1, 'error':'no points found'}

    if rpwn=='yes':
       ftable1=[]
       ctable1=[]

       for q in range(0, len(ftable)):
           fv=ftable[q]
           cv=ctable[q]

           add=True
           for k in fv:
               if k==None:
                  add=False
                  break

           if add:
              for k in cv:
                  if k==None:
                     add=False
                     break

           if add:
              ftable1.append(fv)
              ctable1.append(cv)

       ftable=ftable1
       ctable=ctable1

    if cf!='':
       # Prepare common table from features and characteristics

       lftable=len(ftable)
       lctable=len(ctable)

       if lftable!=lctable:
          return {'return':1, 'error':'length of feature table ('+str(lftable)+') is not the same as length of characteristics table ('+str(lctable)+')'}

       dim=[]
       for q in range(0, lftable): 
           vv=[]
           for v in ftable[q]:
               vv.append(v)
           for v in ctable[q]:
               vv.append(v)
           dim.append(vv)

       # Prepare common keys
       keys=[]
       for q in fkeys:
           keys.append(q)
       for q in ckeys:
           keys.append(q)

       # Prepare temporary CSV file
       ii={'action':'convert_table_to_csv',
           'module_uoa':cfg['module_deps']['experiment'],
           'table':dim,
           'keys':keys,
           'file_name':cf,
           'csv_no_header':'no',
           'csv_separator':';',
           'csv_decimal_mark':'.'
          }
       r=ck.access(ii)
       if r['return']>0: return r

       return {'return':0}

    mmuoa=i['model_module_uoa']
    mn=i['model_name']
    mof=i.get('model_file','')

    # Check if need to crate container (CK entry)
    mduoa=i.get('model_data_uoa','')
    mruoa=i.get('model_repo_uoa','')
    if mduoa!='':
       r=ck.access({'action':'update',
                    'module_uoa':work['self_module_uid'],
                    'data_uoa':mduoa,
                    'repo_uoa':mruoa,
                    'dict':{'model_module_uoa':mmuoa,
                            'model_file':mof,
                            'model_name':mn},
                    'substitute':'yes',
                    'sort_keys':'yes'})
       if r['return']>0: return r
       p=r['path']

       mof=os.path.join(p, mof)

    if mof!='' and os.path.isfile(mof): os.remove(mof)

    # Calling model
    ii={'action':'build',
        'module_uoa':mmuoa,
        'model_name':mn,
        'model_params':i.get('model_params',{}),
        'model_file':mof,
        'features_table': ftable,
        'features_keys': fkeys,
        'features_desc': fdesc,
        'characteristics_table': ctable,
        'characteristics_keys': ckeys,
        'keep_temp_files':ktf,
        'caption':cap,
        'out':o
       }
    r=ck.access(ii)
    if r['return']>0: return r

    mif=r.get('model_input_file','')
    mf=r['model_file']

    if o=='con':
       if ktf=='yes' and mif!='':
         ck.out('Temp model input file '+mif)
       ck.out('Generated model was saved into file '+mf)

    i['out']=o

    return r

##############################################################################
# validate model (universal)

def validate(i):
    """

    Input:  {
              Use data for modeling directly
                (ftable)                                - Feature table
                                                            [ [f1,f2,...], [f1,f2,...], ...]
                (fkeys)                                 - Key names
                                                            [ "f1 name", "f2 name", ...]

                (ctable)                                - Outcome to model
                                                            [ [value], [value], ... ]
                (ckeys)                                 - Key names
                                                            [ "value name" ]
              
              OR select entries:
                (repo_uoa) or (experiment_repo_uoa)     - can be wild cards
                (remote_repo_uoa)                       - if remote access, use this as a remote repo UOA
                (module_uoa) or (experiment_module_uoa) - can be wild cards
                (data_uoa) or (experiment_data_uoa)     - can be wild cards

                (repo_uoa_list)                         - list of repos to search
                (module_uoa_list)                       - list of module to search
                (data_uoa_list)                         - list of data to search

                (search_dict)                           - search dict
                (ignore_case)                           - if 'yes', ignore case when searching

                (features_flat_keys_list)               - list of flat keys to extract from points into table
                                                          (order is important: for example, for plot -> X,Y,Z)
                (features_flat_keys_index)              - add all flat keys starting from this index 
                                                          (for example, ##features#)

                (characteristics_flat_keys_list)        - list of flat keys to extract from points into table
                                                          (order is important: for example, for plot -> X,Y,Z)
                (characteristics_flat_keys_index)       - add all flat keys starting from this index 
                                                          (for example, ##features#)

              Model:
                model_module_uoa                        - model module
                model_name                              - model name
                model_file                              - model file
                (keep_temp_files)                       - if 'yes', keep temp files 
                (remove_points_with_none)               - if 'yes', remote points with None

                (model_data_uoa)                       - record to this data_uoa (container: model_module_uoa)
                (model_repo_uoa)                       - use this repo to record model
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              rmse
              prediction_rate
              observations
              mispredictions
            }

    """

    import copy
    import math
    import os

    o=i.get('out','')
    i['out']=''

    rpwn=i.get('remove_points_with_none','')

    mduoa=i.get('model_data_uoa','')
    mruoa=i.get('model_repo_uoa','')

    if mduoa=='':
       mmuoa=i['model_module_uoa']
       mn=i['model_name']
       mf=i['model_file']
    else:
       r=ck.access({'action':'load',
                    'module_uoa':work['self_module_uid'],
                    'data_uoa':mduoa,
                    'repo_uoa':mruoa})
       if r['return']>0: return r
       p=r['path']
       d=r['dict']

       mmuoa=d['model_module_uoa']
       mn=d['model_name']
       mf=d['model_file']
       
       mf=os.path.join(p,mf)

    mf8=mf+'.model.validation-with-labels.csv'
    if os.path.isfile(mf8): os.remove(mf8)

    ktf=i.get('keep_temp_files','')

    ffkl=i.get('features_flat_keys_list',[])
    fdesc=i.get('features_flat_keys_desc',{})

    ffke=i.get('features_flat_keys_ext','')
    if ffke!='':
       ffkl1=[]
       for q in ffkl:
           q+=ffke
           ffkl1.append(q)
       ffkl=ffkl1

       fdesc1={}
       for q in fdesc:
           fdesc1[q+ffke]=fdesc[q]
       fdesc=fdesc1

    # Get table through experiment module for features
    ftable=i.get('ftable',[])
    fkeys=i.get('fkeys',[])
    mtable=i.get('mtable',[])

    if len(ftable)==0:
       iif=copy.deepcopy(i)
       iif['action']='get'
       iif['module_uoa']=cfg['module_deps']['experiment']
       iif['flat_keys_list']=ffkl
       iif['flat_keys_index']=i.get('features_flat_keys_index','')
       r=ck.access(iif)
       if r['return']>0: return r
       ftable=r['table'].get('0',[])
       fkeys=r['real_keys']

       if len(ftable)==0:
          return {'return':1, 'error':'no points found'}

       mtable=r['mtable'].get('0',[])

    # Get table through experiment module for characteristics
    ctable=i.get('ctable',[])
    ckeys=i.get('ckeys',[])
    if len(ctable)==0:
       iic=copy.deepcopy(i)
       iic['action']='get'
       iic['module_uoa']=cfg['module_deps']['experiment']
       iic['flat_keys_list']=i.get('characteristics_flat_keys_list',[])
       iic['flat_keys_index']=i.get('characteristics_flat_keys_index','')
       r=ck.access(iic)
       if r['return']>0: return r

       ctable=r['table'].get('0',[])
       ckeys=r['real_keys']

    if rpwn=='yes':
       ftable1=[]
       ctable1=[]

       for q in range(0, len(ftable)):
           fv=ftable[q]
           cv=ctable[q]

           add=True
           for k in fv:
               if k==None:
                  add=False
                  break

           if add:
              for k in cv:
                  if k==None:
                     add=False
                     break

           if add:
              ftable1.append(fv)
              ctable1.append(cv)

       ftable=ftable1
       ctable=ctable1

    lctable=len(ctable)
    if lctable==0:
       return {'return':1, 'error':'no points found'}

    # Calling model
    ii={'action':'validate',
        'module_uoa':mmuoa,
        'model_name':mn,
        'model_file':mf,
        'features_table': ftable,
        'features_keys': fkeys,
        'keep_temp_files':ktf,
        'model_params':i.get('model_params',{}),
        'out':o
       }
    r=ck.access(ii)
    if r['return']>0: return r

    pt=r['prediction_table']
    lpt=len(pt)

    lt=r.get('label_table',[])

    if lctable!=lpt:
       return {'return':1, 'error':'length of characteristic table ('+str(lctable)+') is not the same as table with predictions ('+str(lpt)+')'}

    # Get all keys + make header for CVS with leaf labels
    sx='Label:;Original value;predicted value;correct?'
    dkk=[]
    for pp in mtable:
        kk=pp.get('features',{}).get('features',{})
        for a in sorted(kk):
            if a not in dkk: dkk.append(a)
    dkk=sorted(dkk)
    for a in dkk:
        sx+=';'+a
    sx+='\n'

    # Checking model
    s=0.0

    imispredictions=0

    for k in range(0, lctable):
        v=ctable[k][0]
        pv=pt[k][0]

        label='' # only for decision trees
        if len(lt)>0:
           label=lt[k][0]

        if k<len(mtable):
           kk=mtable[k].get('features',{}).get('features',{})

        sdiff=''
        correct=True

        if type(v)==float or type(v)==int:
           if v==0:
              if v!=pv:
                 sdiff='***'
                 imispredictions+=1
                 correct=False
              else:
                 sdiff='0.000'
           else:
              s+=(v-pv)*(v-pv)
              diff=abs(pv-v)/v
              x1=''
              if diff>0.1: #10%
                 x1=' ***'
                 imispredictions+=1
                 correct=False
              sdiff="%7.3f" % diff + x1         

        else:
           if type(v)==bool:
              # hack
              xfloat=True
              try: xpv=float(pv)
              except Exception as e: xfloat=False

              if xfloat:
                 if xpv<0.5: 
                    spv='False ('+("%11.3f" % xpv)+')'
                    pv='False'
                 else: 
                    spv='True ('+("%11.3f" % xpv)+')'
                    pv='True'

              if str(pv)=='True': pt[k][0]=True
              else: pt[k][0]=False

           if str(v)!=str(pv):
              s+=1
              sdiff='***'
              imispredictions+=1
              correct=False

        if type(v)==float:
           sv="%11.3f" % v
           pv=float(pv)
           pt[k][0]=pv
           spv="%11.3f" % pv
        else:
           sv=str(v)
           spv=str(pv)

        sx+=label+';'+sv+';'+spv+';'+str(correct)
        for a in dkk:
            sx+=';'+str(kk.get(a,''))
        sx+='\n'

        if o=='con':
           import json
           x=json.dumps(ftable[k])
           line=str(1+k)+') '+x+' => '+sv+' '+spv+' '+sdiff
           ck.out(line)

    rmse=math.sqrt(s/lctable)
    rate=float(lctable-imispredictions)/float(lctable)

    r=ck.save_text_file({'text_file':mf8, 'string':sx})
    if r['return']>0: return r

    if o=='con':
       ck.out('')      

       ck.out('Model RMSE =      '+str(rmse))
       ck.out('Prediction rate = '+("%4.3f" % (rate*100))+'%')
       ck.out('Mispredictions =  '+str(imispredictions)+' out of '+str(lctable))

    # Visualize?
    if i.get('visualize','')=='yes':
       if o=='con':
          ck.out('')
          ck.out('Attempting to visualize ...')

       table={"0":[], "1":[]}
       for k in range(0, lctable):
           table["0"].append([0, ctable[k][0]])
           table["1"].append([0, pt[k][0]])

       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],
           'table':table}
       iig=i.get('graph_params',{})
       ii.update(iig)
       r=ck.access(ii)
       if r['return']>0: return r

    i['out']=o
    return {'return':0, 'rmse':rmse, 'prediction_rate':rate, 'observations':lctable, 'mispredictions':imispredictions}

##############################################################################
# convert table to CSV

def convert_to_csv(i):
    """
    Input:  {
              Input from 'build' function

              csv_file - CSV file to record
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    return build(i)
